@@ -97,7 +97,7 @@ static const struct got_error *gotweb_get_clone_url(char **, struct server *,
 static void gotweb_free_querystring(struct querystring *);
 static void gotweb_free_repo_dir(struct repo_dir *);
 
-struct server *gotweb_get_server(uint8_t *, uint8_t *);
+struct server *gotweb_get_server(const char *);
 
 static int
 gotweb_reply(struct request *c, int status, const char *ctype,
@@ -161,7 +161,7 @@ gotweb_process_request(struct request *c)
 	if (c->sock->client_status == CLIENT_DISCONNECT)
 		return;
 	/* get the gotwebd server */
-	srv = gotweb_get_server(c->server_name, c->http_host);
+	srv = gotweb_get_server(c->server_name);
 	if (srv == NULL) {
 		log_warnx("%s: error server is NULL", __func__);
 		goto err;
@@ -189,7 +189,7 @@ gotweb_process_request(struct request *c)
 	if (qs->action == BLAME || qs->action == BLOB ||
 	    qs->action == BLOBRAW || qs->action == DIFF) {
 		if (qs->commit == NULL) {
-			error = got_error(GOT_ERR_QUERYSTRING);
+			error = got_error(GOT_ERR_BAD_QUERYSTRING);
 			goto err;
 		}
 	}
@@ -267,6 +267,9 @@ gotweb_process_request(struct request *c)
 		}
 		return;
 	case BRIEFS:
+		error = got_get_repo_commits(c, srv->max_commits_display);
+		if (error)
+			goto err;
 		if (gotweb_reply(c, 200, "text/html", NULL) == -1)
 			return;
 		gotweb_render_page(c->tp, gotweb_render_briefs);
@@ -326,6 +329,9 @@ gotweb_process_request(struct request *c)
 			    error->msg);
 			goto err;
 		}
+		error = got_get_repo_commits(c, D_MAXSLCOMMDISP);
+		if (error)
+			goto err;
 		qs->action = TAGS;
 		error = got_get_repo_tags(c, D_MAXSLCOMMDISP);
 		if (error) {
@@ -386,28 +392,18 @@ err:
 }
 
 struct server *
-gotweb_get_server(uint8_t *server_name, uint8_t *subdomain)
+gotweb_get_server(const char *server_name)
 {
-	struct server *srv = NULL;
+	struct server *srv;
 
 	/* check against the server name first */
-	if (strlen(server_name) > 0)
+	if (*server_name != '\0')
 		TAILQ_FOREACH(srv, &gotwebd_env->servers, entry)
 			if (strcmp(srv->name, server_name) == 0)
-				goto done;
+				return srv;
 
-	/* check against subdomain second */
-	if (strlen(subdomain) > 0)
-		TAILQ_FOREACH(srv, &gotwebd_env->servers, entry)
-			if (strcmp(srv->name, subdomain) == 0)
-				goto done;
-
-	/* if those fail, send first server */
-	TAILQ_FOREACH(srv, &gotwebd_env->servers, entry)
-		if (srv != NULL)
-			break;
-done:
-	return srv;
+	/* otherwise, use the first server */
+	return TAILQ_FIRST(&gotwebd_env->servers);
 };
 
 const struct got_error *
@@ -613,7 +609,7 @@ qa_found:
 			}
 			break;
 		case INDEX_PAGE:
-			if (strlen(value) == 0)
+			if (*value == '\0')
 				break;
 			(*qs)->index_page = strtonum(value, INT64_MIN,
 			    INT64_MAX, &errstr);
@@ -634,7 +630,7 @@ qa_found:
 			}
 			break;
 		case PAGE:
-			if (strlen(value) == 0)
+			if (*value == '\0')
 				break;
 			(*qs)->page = strtonum(value, INT64_MIN,
 			    INT64_MAX, &errstr);
